import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import java.util.*;
import ij.measure.*;

/** This plugin levels the surface of tilted images

	Gary Chinga 020506

 */

public class Surface_leveling implements PlugInFilter {

	ImagePlus imp;
    	ResultsTable rt = new ResultsTable();
	boolean canceled = false;
	boolean coordinates = false;
	boolean regressionDetails = false;
	double pixelSize;
	int n;
	float Ex1, Ex1x1, Ex2x2, Ex1x2, Ey, Ex2;
	float x1Mean, x2Mean, yMean, Eyy, Ex1y, Ex2y;
	float Sx1y, Sx2y, Sx1x2, Sx1x1, Sx2x2, b1, b2, b0;



	public int setup(String arg, ImagePlus imp) {
		if (IJ.versionLessThan("1.17j"))
			return DONE;
		this.imp = imp;
		return DOES_32;
	}

	public void run(ImageProcessor ip) {

		int nSlices = imp.getStackSize();
		ImageStack stack = imp.getStack();
		int w = stack.getWidth();
		int h = stack.getHeight();

		//Create new stacks for the output.
       	ImageStack levelledStack = new ImageStack(w,h);

       	//Go through the slices of the input stack.
		for (int i = 0; i < nSlices; i++){

			IJ.showStatus("Please wait ...  Leveling surface: " + i);

			//Get a slice and make copies of the pixel arrays.
 			ImageProcessor ipSlice = stack.getProcessor(i+1);
 			float[] tempStack = (float[])ipSlice.getPixelsCopy();

 			//Put the pixel arrays into image processors for the convolutions.
 			ImageProcessor ipTempStack= new FloatProcessor(w,h);
 			ipTempStack.setPixels(tempStack);

			//Make places for the results.
			FloatProcessor ipTemp = new FloatProcessor(w,h);
			float[] temp = (float[])ipTemp.getPixels();

			//Runs the same routines 4 times
			for (int it=0; it<4;it++){
				n=0;
				Ex1=0; Ex1x1=0; Ex2x2=0; Ex1x2=0; Ey=0; Ex2=0;
				x1Mean = 0; x2Mean=0; yMean = 0; Eyy = 0; Ex1y = 0; Ex2y=0;
				Sx1y = 0; Sx2y=0; Sx1x2 = 0; Sx1x1=0; Sx2x2=0; b1 = 0; b2 = 0; b0 = 0;

				//Go through the pixel values in each slice (image)
				//x1 og x2 are the pixel coordinates, y are the topography values
				for (int x2= 0; x2 < h; x2++){
					for (int x1 = 0; x1 < w; x1++){
						n++;
						int index = x1 + w*x2;
						//int index2 = x2 + h*x1;
						float y = tempStack[index];

						Ex1 = Ex1 + x1;			//Sums x, y...
						Ex2 = Ex2 + x2;
						Ey = Ey + y;
						Ex1x1 = Ex1x1 + x1*x1;	//Sums of squares and
						Ex2x2 = Ex2x2 + x2*x2;
						Ex1x2 = Ex1x2 + x1*x2;	//cross products of variables
						Ex1y = Ex1y + x1*y;
						Ex2y = Ex2y + x2*y;
					}
				}

				yMean = Ey / n;
				x1Mean = Ex1 / n;
				x2Mean = Ex2 / n;

				Sx1x2 = Ex1x2 - (n*(x1Mean)*(x2Mean));
				Sx1x1 = Ex1x1 - (n*(x1Mean)*(x1Mean));
				Sx2x2 = Ex2x2 - (n*(x2Mean)*(x2Mean));

				Sx1y = Ex1y - (n*(x1Mean)*(yMean));
				Sx2y = Ex2y - (n*(x2Mean)*(yMean));

				//calculate the values of the least square estimates
				b1 = (((Sx2x2*Sx1y)+(Sx1x2*Sx2y))/((Sx1x1*Sx2x2)+(Sx1x2*Sx1x2)));
				b2 = ((Sx2y-(b1*Sx1x2))/(Sx2x2));
				b0 = yMean - b1*x1Mean - b2*x2Mean;

				for (int x2 = 0; x2 < h; x2++){
					for (int x1 = 0; x1 < w; x1++){
						int index = x1 + w*x2;
						float y = tempStack[index];

						//regression model applied to each pixel in the image
						temp[index] = (float) (y-(b0+b1*x1 + b2*x2));
						tempStack[index] = temp[index];
					}
				}

				//Put the results in the destination stacks.
				if (it==3){
					ipTemp.setMinAndMax(0,0);
					levelledStack.addSlice("LevelledImages",ipTemp);
					}

					//Records the results for the first levelling
					if (it==0) {
							//Write the results in the result table
						if (regressionDetails){
							rt.incrementCounter();
							rt.addValue("n", n);
							rt.addValue("yMean", yMean);
							rt.addValue("x1Mean", x1Mean);
							rt.addValue("x2Mean", x2Mean);
							rt.addValue("b0", b0);
							rt.addValue("b1", b1);
							rt.addValue("b2", b2);
							rt.addValue("Sx1y", Sx1y);
							rt.addValue("Sx2y", Sx2y);
							rt.addValue("Sx1x1", Sx1x1);
							rt.addValue("Sx2x2", Sx2x2);
							rt.addValue("Sx1x2", Sx1x2);
							if (i==0)
								IJ.setColumnHeadings(rt.getColumnHeadings());
							IJ.write(rt.getRowAsString(rt.getCounter()-1));
						}
					} //if
				} //iterations
		    }	//slice

			// Create new images using the new stacks.
			ImagePlus impLevelled = new ImagePlus("LevelledImages",levelledStack);
			impLevelled.setStack(null,levelledStack);
			// Display the new stacks.
			impLevelled.show();

		} //for
	}

