import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.plugin.filter.GaussianBlur;
import ij.process.*;
import ij.gui.*;
import ij.measure.*;
import java.awt.*;
import java.util.*;
import java.awt.event.*;


/** 	This plugin separates the surface topography into a roughness and a waviness image.
 
		Gary Chinga 020726
*/

public class Waveness_Roughness implements PlugInFilter {

    static int pixelSize = 1;
    static int sampleSize = 4;

    boolean crop = false;
    boolean canceled = false;
    int r,ss;
    double ps;
    ImagePlus imp;
    int w,h,c;
    public int setup(String arg, ImagePlus imp) {
        if (IJ.versionLessThan("1.26"))
            return DONE;
        this.imp = imp;
        return DOES_ALL-DOES_RGB+NO_UNDO;
    }

    public void run(ImageProcessor ip) {

        ImageStack stack = imp.getStack();

        w = stack.getWidth();
        h = stack.getHeight();
        if (h<w)
            c=h;
        else
            c=w;

        int nSlices = imp.getStackSize();

        GaussianBlur gb = new GaussianBlur();	//Construct a GaussianBlur filter

        Calibration cal = imp.getCalibration();
        ps = cal.pixelWidth;
        getGaussianRadius();
        if (canceled)
            return;

        //Divide the entered radius with the pixel size - CALIBRATION
        //Calculate also the crop dimensions
        r=(int)(ss/(2*ps));
        int xx=r, yy=r;
        int ww=w-2*r, hh=h-2*r;

        //Create new stacks for the output.
        ImageStack imsWaveness = new ImageStack(w,h);
        ImageStack imsRoughness = new ImageStack(w,h);

        for (int i = 0; i < nSlices; i++) {
            //IJ.showProgress((double)i/nSlices);
            IJ.showStatus("a: "+i+"/"+nSlices);
            imp.setSlice(i+1);		//Set slice in stack

            ip = ip.convertToFloat();
            float [] pixels = (float[]) ip.getPixels();
            ImageProcessor ipSlice = stack.getProcessor(i+1);
            ipSlice = ipSlice.convertToFloat();
            float[] waveness = (float[])ipSlice.getPixelsCopy();

            //Put the pixel arrays into image processors for the convolutions.
            ImageProcessor ipWaveness= new FloatProcessor(w,h);
            ipWaveness.setPixels(waveness);

            //Detect the surface waveness
            gb.blur(ipWaveness,r);

            //Make places for the results.
            FloatProcessor ipTempRoughness = new FloatProcessor(w,h);
            float[] tempRoughness = (float[])ipTempRoughness.getPixels();

            //Put the pixel arrays into image processors for the convolutions.
            //ImageProcessor ipRoughness= new FloatProcessor(w,h);
            //ipRoughness.setPixels(tempRoughness);

            //Extract the roughness between the stacks
            for (int y = 0; y < h; y++) {
                for (int x = 0; x < w; x++) {
                    int index = x + w*y;
                    //Perform image substraction
                    tempRoughness[index] = pixels[index]-waveness[index];
                }
            }
            //Put the results in the destination stacks.
            ipWaveness.setMinAndMax(0,0);
            imsWaveness.addSlice("Waviness",ipWaveness);
            //ipRoughness.setMinAndMax(0,0);
            imsRoughness.addSlice("Roughness",ipTempRoughness);
        }

        // Create new images using the new stacks.
        ImagePlus impWaveness = new ImagePlus("Waviness",imsWaveness);
        ImagePlus impRoughness = new ImagePlus("Roughness",imsRoughness);
        impWaveness.setStack(null,imsWaveness);
        impRoughness.setStack(null,imsRoughness);

        //Display
        impWaveness.show();
        IJ.run("Fire");
        if (crop) {
            IJ.makeRectangle(xx,yy,ww,hh);
            IJ.run("Crop");
        }
        impRoughness.show();
        IJ.run("Fire");
        if (crop) {
            IJ.makeRectangle(xx,yy,ww,hh);
            IJ.run("Crop");
        }
        //IJ.showProgress(1.0);
        IJ.run("Tile");

        //imp.setStack(null,stack);			//Update the image for the stack
    }

    void getGaussianRadius() {
        GenericDialog gd = new GenericDialog("Filter details...");
        String units = imp.getCalibration().getUnits();
        gd.addNumericField(" Cut-off Value ("+units+"): ", sampleSize, 0);
        if (ps==1)
            units = imp.getCalibration().getUnit();
        int decimalPlaces = (int)ps==ps?0:2;
        if (!units.startsWith("pixel"))
            gd.addMessage("        (Pixel size = "+IJ.d2s(ps,decimalPlaces)+" "+units+")");
        gd.addCheckbox("Crop image depending on cut-off value ", crop);
        gd.showDialog();
        if (gd.wasCanceled()) {
            canceled = true;
            return;
        }
        ps = pixelSize;
        ss = (int)gd.getNextNumber();
        crop = gd.getNextBoolean();
        if (ss>(c*ps/2) || ss<(4*ps)) {
            canceled = true;
            IJ.showMessage("Min and Max recommened radius are "+(4*ps)+" and "+(c*ps/2)+" respectively");
            return;
        }
    }
}
